//	Copyright © 2011 The Go Authors.  All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// This is copied from architecture dependent header files to facilitate
// building ARM support on x86 and x86 support on ARM. Because ARM related
// headers do not exist on Mac OS X, and vice versa.
// XXX(minux): how to keep this file up-to-date?

// ptrace, copied from /usr/include/sys/ptrace.h, this file doesn't exist in iOS SDK
#define	PT_TRACE_ME	0	/* child declares it's being traced */
#define	PT_CONTINUE	7	/* continue the child */
#define	PT_KILL		8	/* kill the child process */
int	ptrace(int _request, pid_t _pid, caddr_t _addr, int _data);

// ARM, copied from /usr/include/mach/arm/{thread_status,_structs}.h
#ifndef ARM_THREAD_STATE
#define ARM_THREAD_STATE	1

#define ARM_THREAD_STATE_COUNT ((mach_msg_type_number_t) \
   (sizeof (arm_thread_state_t)/sizeof(uint32_t)))
#define _STRUCT_ARM_THREAD_STATE	struct arm_thread_state
typedef _STRUCT_ARM_THREAD_STATE
{
	__uint32_t	r[13];	/* General purpose register r0-r12 */
	__uint32_t	sp;		/* Stack pointer r13 */
	__uint32_t	lr;		/* Link register r14 */
	__uint32_t	pc;		/* Program counter r15 */
	__uint32_t	cpsr;		/* Current program status register */
} arm_thread_state_t;
#endif

// x86-32 and x86-64, copied from /usr/include/mach/i386/{thread_status,_structs}.h
#ifndef x86_THREAD_STATE32
#define x86_THREAD_STATE32		1
#define x86_THREAD_STATE64		4

#define x86_THREAD_STATE32_COUNT	((mach_msg_type_number_t) \
    ( sizeof (x86_thread_state32_t) / sizeof (int) ))
#define	_STRUCT_X86_THREAD_STATE32	struct x86_thread_state32
typedef _STRUCT_X86_THREAD_STATE32
{
    unsigned int	eax;
    unsigned int	ebx;
    unsigned int	ecx;
    unsigned int	edx;
    unsigned int	edi;
    unsigned int	esi;
    unsigned int	ebp;
    unsigned int	esp;
    unsigned int	ss;
    unsigned int	eflags;
    unsigned int	eip;
    unsigned int	cs;
    unsigned int	ds;
    unsigned int	es;
    unsigned int	fs;
    unsigned int	gs;
} x86_thread_state32_t;

#define x86_THREAD_STATE64_COUNT	((mach_msg_type_number_t) \
    ( sizeof (x86_thread_state64_t) / sizeof (int) ))
#define	_STRUCT_X86_THREAD_STATE64	struct x86_thread_state64
typedef _STRUCT_X86_THREAD_STATE64
{
	__uint64_t	rax;
	__uint64_t	rbx;
	__uint64_t	rcx;
	__uint64_t	rdx;
	__uint64_t	rdi;
	__uint64_t	rsi;
	__uint64_t	rbp;
	__uint64_t	rsp;
	__uint64_t	r8;
	__uint64_t	r9;
	__uint64_t	r10;
	__uint64_t	r11;
	__uint64_t	r12;
	__uint64_t	r13;
	__uint64_t	r14;
	__uint64_t	r15;
	__uint64_t	rip;
	__uint64_t	rflags;
	__uint64_t	cs;
	__uint64_t	fs;
	__uint64_t	gs;
} x86_thread_state64_t;
#endif
